"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var typescript_1 = __importDefault(require("typescript"));
var File_1 = require("./components/File");
var Parameter_1 = require("./components/Parameter");
var Class_1 = require("./components/Class");
var Method_1 = require("./components/Method");
function generateDocumentation(fileNames, options) {
    if (options === void 0) { options = typescript_1.default.getDefaultCompilerOptions(); }
    var program = typescript_1.default.createProgram(fileNames, options);
    var checker = program.getTypeChecker();
    var result = [];
    program.getSourceFiles()
        .forEach(function (sourceFile) {
        if (!sourceFile.isDeclarationFile) {
            var file = getFile(sourceFile, checker);
            if (file !== undefined) {
                result.push(file);
            }
        }
    });
    return output;
}
exports.generateDocumentation = generateDocumentation;
function getFile(sourceFile, checker) {
    var file = new File_1.File();
    typescript_1.default.forEachChild(sourceFile, function (node) {
        if (!isNodeExported(node)) {
            return;
        }
        if (node.kind === typescript_1.default.SyntaxKind.ClassDeclaration) {
            var currentNode = node;
            if (currentNode.name === undefined) {
                return;
            }
            var symbol = checker.getSymbolAtLocation(currentNode.name);
            if (symbol === undefined) {
                return;
            }
            file.classes.push(serializeClass(symbol, checker));
        }
        else if (node.kind === typescript_1.default.SyntaxKind.InterfaceDeclaration) {
            var currentNode = node;
            var symbol = checker.getSymbolAtLocation(currentNode.name);
            if (symbol === undefined) {
                return;
            }
            file.interfaces.push(serializeInterface(symbol));
        }
        else if (node.kind === typescript_1.default.SyntaxKind.ModuleDeclaration) {
            return;
        }
        else if (node.kind === typescript_1.default.SyntaxKind.EnumDeclaration) {
            var currentNode = node;
            var symbol = checker.getSymbolAtLocation(currentNode.name);
            if (symbol === undefined) {
                return;
            }
            file.enums.push(serializeEnum(symbol));
            return;
        }
    });
    return file;
}
function isOptional(declarations) {
    for (var _i = 0, declarations_1 = declarations; _i < declarations_1.length; _i++) {
        var declaration = declarations_1[_i];
        if (!(declaration.questionToken === undefined &&
            declaration.initializer === undefined)) {
            return true;
        }
    }
    return false;
}
function serializeParameter(symbol) {
    var result = new Parameter_1.Parameter();
    var declarations = symbol.getDeclarations();
    var questionToken = false;
    if (declarations !== undefined) {
        result.isOptional = isOptional(declarations);
    }
    result.name = symbol.getName();
    result.type = checker.typeToString(checker.getTypeOfSymbolAtLocation(symbol, symbol.valueDeclaration));
    return result;
}
function serializeMember(memberSymbol, memberDeclaration) {
    var result = {
        keyword: getMemberKeyword(memberDeclaration),
        modifierType: getMemberModifierType(memberDeclaration),
        name: memberSymbol.getName(),
        parameters: [],
        returnType: getMemberReturnType(memberSymbol, memberDeclaration),
        type: getMemberType(memberDeclaration),
        questionToken: getQuestionToken(memberDeclaration),
        constraint: getConstraint(memberDeclaration)
    };
    if (result.type === MEMBER_TYPE.CONSTRUCTOR) {
        return;
    }
    if (result.type === MEMBER_TYPE.METHOD) {
        var methodSignature = checker.getSignatureFromDeclaration(memberDeclaration);
        if (methodSignature === undefined) {
            return result;
        }
        return __assign({}, result, serializeSignature(methodSignature));
    }
    return result;
}
function getMemberReturnType(memberSymbol, memberDeclaration) {
    var memberType = getMemberType(memberDeclaration);
    if (memberType === MEMBER_TYPE.ENUM || memberType === MEMBER_TYPE.CONSTRUCTOR) {
        return '';
    }
    return checker.typeToString(checker.getTypeOfSymbolAtLocation(memberSymbol, memberDeclaration));
}
function getQuestionToken(memberDeclaration) {
    var propertyDeclaration = memberDeclaration;
    return propertyDeclaration.questionToken !== undefined;
}
function getConstraint(memberDeclaration) {
    var effectiveConstraint = typescript_1.default.getEffectiveConstraintOfTypeParameter(memberDeclaration);
    if (effectiveConstraint === undefined) {
        return;
    }
    return checker.typeToString(checker.getTypeFromTypeNode(effectiveConstraint));
}
function getMemberType(memberDeclaration) {
    switch (memberDeclaration.kind) {
        case typescript_1.default.SyntaxKind.PropertyDeclaration:
        case typescript_1.default.SyntaxKind.PropertySignature:
        case typescript_1.default.SyntaxKind.Parameter:
        case typescript_1.default.SyntaxKind.GetAccessor:
        case typescript_1.default.SyntaxKind.SetAccessor:
            return MEMBER_TYPE.PROPERTY;
        case typescript_1.default.SyntaxKind.MethodDeclaration:
        case typescript_1.default.SyntaxKind.MethodSignature:
            return MEMBER_TYPE.METHOD;
        case typescript_1.default.SyntaxKind.Constructor:
        case typescript_1.default.SyntaxKind.ConstructSignature:
            return MEMBER_TYPE.CONSTRUCTOR;
        case typescript_1.default.SyntaxKind.IndexSignature:
            return MEMBER_TYPE.INDEX;
        case typescript_1.default.SyntaxKind.TypeParameter:
            return MEMBER_TYPE.PARAMETER;
        case typescript_1.default.SyntaxKind.EnumMember:
        case typescript_1.default.SyntaxKind.EnumDeclaration:
            return MEMBER_TYPE.ENUM;
        default:
            throw new Error('unable to determine member type');
    }
}
function getModifierType(modifiers) {
    var modifierType = MODIFIER_TYPE.PUBLIC;
    modifiers.some(function (modifier) {
        if (modifier.kind === typescript_1.default.SyntaxKind.PrivateKeyword) {
            modifierType = MODIFIER_TYPE.PRIVATE;
            return true;
        }
        if (modifier.kind === typescript_1.default.SyntaxKind.PublicKeyword) {
            modifierType = MODIFIER_TYPE.PUBLIC;
            return true;
        }
        if (modifier.kind === typescript_1.default.SyntaxKind.ProtectedKeyword) {
            modifierType = MODIFIER_TYPE.PROTECTED;
            return true;
        }
        return false;
    });
    return modifierType;
}
function getMemberModifierType(memberDeclaration) {
    var memberModifiers = memberDeclaration.modifiers;
    if (memberModifiers === undefined ||
        memberModifiers.length === 0) {
        return MODIFIER_TYPE.PUBLIC;
    }
    return getModifierType(memberModifiers);
}
function isAbstract(memberDeclaration) {
    var memberModifiers = memberDeclaration.modifiers;
    if (memberModifiers !== undefined) {
        for (var _i = 0, memberModifiers_1 = memberModifiers; _i < memberModifiers_1.length; _i++) {
            var memberModifier = memberModifiers_1[_i];
            if (memberModifier.kind === typescript_1.default.SyntaxKind.AbstractKeyword) {
                return true;
            }
        }
    }
    return false;
}
function isStatic(memberDeclaration) {
    var memberModifiers = memberDeclaration.modifiers;
    if (memberModifiers !== undefined) {
        for (var _i = 0, memberModifiers_2 = memberModifiers; _i < memberModifiers_2.length; _i++) {
            var memberModifier = memberModifiers_2[_i];
            if (memberModifier.kind === typescript_1.default.SyntaxKind.StaticKeyword) {
                return true;
            }
        }
    }
    return false;
}
function getMemberKeyword(memberDeclaration) {
    var memberKeyword;
    var memberModifiers = memberDeclaration.modifiers;
    if (memberModifiers === undefined ||
        memberModifiers.length === 0) {
        return memberKeyword;
    }
    memberModifiers.some(function (memberModifier) {
        if (memberModifier.kind === typescript_1.default.SyntaxKind.AbstractKeyword) {
            memberKeyword = CLASS_MEMBER_KEYWORD.ABSTRACT;
            return true;
        }
        if (memberModifier.kind === typescript_1.default.SyntaxKind.StaticKeyword) {
            memberKeyword = CLASS_MEMBER_KEYWORD.STATIC;
            return true;
        }
        return false;
    });
    return memberKeyword;
}
function getExtendsHeritageClauseName(heritageClause) {
    return heritageClause.types[0].expression.text;
}
function getInterfaceName(nodeObject) {
    return nodeObject.expression.text;
}
function getImplementsHeritageClauseNames(heritageClause) {
    return heritageClause.types.map(getInterfaceName);
}
function serializeEnum(enumSymbol) {
    return __assign({}, serializeInterface(enumSymbol), { structure: STRUCTURE.ENUM });
}
function serializeInterface(interfaceSymbol) {
    var serializedInterface = __assign({}, serializeSymbol(interfaceSymbol), { members: [], structure: STRUCTURE.INTERFACE });
    if (interfaceSymbol.members !== undefined) {
        interfaceSymbol.members.forEach(handleInterfaceMemberSymbols);
    }
    if (interfaceSymbol.exports !== undefined) {
        interfaceSymbol.exports.forEach(handleInterfaceMemberSymbols);
    }
    if (interfaceSymbol.globalExports !== undefined) {
        interfaceSymbol.globalExports.forEach(handleInterfaceMemberSymbols);
    }
    var interfaceDeclarations = interfaceSymbol.getDeclarations();
    if (interfaceDeclarations === undefined) {
        return serializedInterface;
    }
    interfaceDeclarations.forEach(function (interfaceDeclaration) {
        var heritageClauses = interfaceDeclaration.heritageClauses;
        if (heritageClauses === undefined) {
            return;
        }
        heritageClauses.forEach(function (heritageClause) {
            if (heritageClause.token === typescript_1.default.SyntaxKind.ExtendsKeyword) {
                serializedInterface.extends = getExtendsHeritageClauseName(heritageClause);
                return;
            }
            if (heritageClause.token === typescript_1.default.SyntaxKind.ImplementsKeyword) {
                serializedInterface.implements = getImplementsHeritageClauseNames(heritageClause);
                return;
            }
            throw new Error('unsupported heritage clause');
        });
    });
    return serializedInterface;
    function handleInterfaceMemberSymbols(memberSymbol) {
        var memberDeclarations = memberSymbol.getDeclarations();
        if (memberDeclarations === undefined) {
            return;
        }
        memberDeclarations.forEach(function (memberDeclaration) {
            var serializedMember = serializeMember(memberSymbol, memberDeclaration);
            if (serializedMember === undefined) {
                return;
            }
            if (serializedMember.type === MEMBER_TYPE.PARAMETER) {
                if (serializedInterface.parameters === undefined) {
                    serializedInterface.parameters = [];
                }
                serializedInterface.parameters.push(serializedMember);
                return;
            }
            serializedInterface.members.push(serializedMember);
        });
    }
}
function serializeMethods(memberSymbols, checker) {
    var result = [];
    for (var _i = 0, memberSymbols_1 = memberSymbols; _i < memberSymbols_1.length; _i++) {
        var memberSymbol = memberSymbols_1[_i];
        var memberDeclarations = memberSymbol.getDeclarations();
        if (memberDeclarations === undefined) {
            continue;
        }
        memberDeclarations.forEach(function (memberDeclaration) {
            var method = serializeMethod(memberDeclaration, checker);
            if (method) {
                result.methods.push(method);
            }
        });
    }
    return result;
}
function serializeClass(classSymbol, checker) {
    var result = new Class_1.Class();
    var classDeclaration = classSymbol.getDeclarations();
    result.name = classSymbol.getName();
    if (classDeclaration !== undefined && classDeclaration.length > 0) {
        result.isStatic = isStatic(classDeclaration[classDeclaration.length - 1]);
        result.isAbstract = isAbstract(classDeclaration[classDeclaration.length - 1]);
    }
    if (classSymbol.members !== undefined) {
        result.methods = serializeMethods(classSymbol.members, checker);
    }
    var constructorType = checker.getTypeOfSymbolAtLocation(classSymbol, classSymbol.valueDeclaration);
    if (constructorType) {
        result.constructorMethods = constructorType.getConstructSignatures().map(function (constructorMethod) { return serializeMethod(constructorMethod, checker); });
    }
    return result;
}
function serializeMethod(signature, checker) {
    var result = new Method_1.Method();
    result.returnType = checker.typeToString(signature.getReturnType());
    result.parameters = signature.parameters.map(serializeParameter);
    return result;
}
function isNodeExported(node) {
    return (node.flags & typescript_1.default.ModifierFlags.Export) !== 0 ||
        node.parent.kind === typescript_1.default.SyntaxKind.SourceFile;
}
