#!/usr/bin/env node
"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var commander_1 = __importDefault(require("commander"));
var fs_1 = __importDefault(require("fs"));
var glob_1 = __importDefault(require("glob"));
var os_1 = __importDefault(require("os"));
var path_1 = __importDefault(require("path"));
var typescript_1 = __importDefault(require("typescript"));
var convertToPlant_1 = require("./convertToPlant");
var generateDocumentation_1 = require("./generateDocumentation");
commander_1.default
    .version('2.1.3')
    .usage('[options]')
    .option('-i, --input <path>', 'Define the path of the Typescript file')
    .option('-o, --output <path>', 'Define the path of the output file. If not defined, it\'ll output on the STDOUT')
    .option('-p, --project <path>', 'Compile a project given a valid configuration file.' +
    ' The argument can be a file path to a valid JSON configuration file,' +
    ' or a directory path to a directory containing a tsconfig.json file.')
    .option('-C, --compositions', 'Create not heritage compositions')
    .option('-I, --only-interfaces', 'Only output interfaces')
    .parse(process.argv);
if (!commander_1.default.input) {
    console.error('Missing input file');
    process.exit(1);
}
glob_1.default(commander_1.default.input, {}, function (err, matches) {
    if (err !== null) {
        console.error(err);
        return;
    }
    var tsConfigFile = findTsConfigFile(commander_1.default.input, commander_1.default.tsconfig);
    var output = convertToPlant_1.convertToPlant(generateDocumentation_1.generateDocumentation(matches, getCompilerOptions(tsConfigFile)), {
        compositions: commander_1.default.compositions,
        onlyInterfaces: commander_1.default.onlyInterfaces
    });
    if (commander_1.default.output === undefined) {
        console.log(output);
        return;
    }
    fs_1.default.writeFile(commander_1.default.output, output, function (errNoException) {
        if (errNoException !== null) {
            console.error(errNoException);
            return;
        }
        console.log('The file was saved!');
    });
});
function findTsConfigFile(inputPath, tsConfigPath) {
    if (tsConfigPath !== undefined) {
        var tsConfigStats = fs_1.default.statSync(tsConfigPath);
        if (tsConfigStats.isFile()) {
            return tsConfigPath;
        }
        if (tsConfigStats.isDirectory()) {
            var tsConfigFilePath = path_1.default.resolve(tsConfigPath, 'tsconfig.json');
            if (fs_1.default.existsSync(tsConfigFilePath)) {
                return tsConfigFilePath;
            }
        }
    }
    var localTsConfigFile = path_1.default.resolve(path_1.default.dirname(inputPath), 'tsconfig.json');
    if (fs_1.default.existsSync(localTsConfigFile)) {
        return localTsConfigFile;
    }
    var cwdTsConfigFile = path_1.default.resolve(process.cwd(), 'tsconfig.json');
    if (fs_1.default.existsSync(cwdTsConfigFile)) {
        return cwdTsConfigFile;
    }
}
function getCompilerOptions(tsConfigFilePath) {
    if (tsConfigFilePath === undefined) {
        return typescript_1.default.getDefaultCompilerOptions();
    }
    var reader = function (filePath) { return fs_1.default.readFileSync(filePath, 'utf8'); };
    var configFile = typescript_1.default.readConfigFile(tsConfigFilePath, reader);
    if (configFile.error !== undefined && configFile.error.category === typescript_1.default.DiagnosticCategory.Error) {
        throw new Error("unable to read tsconfig.json file at: " + tsConfigFilePath + ".\n             Error: " + typescript_1.default.flattenDiagnosticMessageText(configFile.error.messageText, os_1.default.EOL));
    }
    else if (configFile.config === undefined) {
        throw new Error("unable to read tsconfig.json file at: " + tsConfigFilePath + ".");
    }
    var convertedCompilerOptions = typescript_1.default.convertCompilerOptionsFromJson(configFile.config.compilerOptions, path_1.default.dirname(tsConfigFilePath));
    if (convertedCompilerOptions.errors.length > 0) {
        convertedCompilerOptions.errors.forEach(function (error) {
            if (error.category === typescript_1.default.DiagnosticCategory.Error) {
                throw new Error("unable to read tsconfig.json file at: " + tsConfigFilePath + ".\n                Error: " + typescript_1.default.flattenDiagnosticMessageText(error.messageText, os_1.default.EOL));
            }
        });
    }
    return convertedCompilerOptions.options;
}
